/* bufstream.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "circularBuffer.h"
#include "mainProgram.h"

#include <algorithm>
#include <iostream.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>

static const int default_buffer_kbytes = 64;

class BufstreamMain : public MainProgram {
public:
  BufstreamMain();

  virtual void run();

protected:
  virtual void handle_opts(const string &opt, const string &arg);
  virtual void help();
  virtual void describe_options();
  virtual void describe_args();

  int _buffer_size;
  int _fill_size;
  bool _verbose;
};

BufstreamMain::
BufstreamMain() : MainProgram("bufstream") {
  define_opt("b", true);
  define_opt("f", true);
  define_opt("v", false);

  _buffer_size = default_buffer_kbytes;
  _fill_size = -1;
  _verbose = false;
}

void BufstreamMain::
run() {
  CircularBuffer cb(_buffer_size * 1024);

  int wait_for_fill = 
    (_fill_size < 0 || _fill_size > _buffer_size) 
    ? _buffer_size * 1024 : _fill_size * 1024;
  bool output_ok = false;
  
  // Now repeatedly fill the buffer from stdin while we drain it to
  // stdout.  We use select() to drive the loop here, since we're
  // don't want to be blocking on input while the output is waiting,
  // and vice-versa.

  // Set both file descriptors to use non-blocking I/O so we can take
  // the greatest advantage of select().
  int result = fcntl(STDIN_FILENO, F_SETFL, O_NONBLOCK);
  if (result < 0) {
    perror("fcntl");
  }
  result = fcntl(STDOUT_FILENO, F_SETFL, O_NONBLOCK);
  if (result < 0) {
    perror("fcntl");
  }
  

  while (true) {
    fd_set in_fds;
    FD_ZERO(&in_fds);
    if (cb.get_current_size() < cb.get_max_size()) {
      assert(cb.get_write_size() > 0);
      // We only care whether input is available if there's room in the
      // buffer.
      FD_SET(STDIN_FILENO, &in_fds);
    }

    if (cb.get_current_size() >= wait_for_fill) {
      // If we've achieved our minimum fill, start sending the output.
      output_ok = true;
    }

    fd_set out_fds;
    FD_ZERO(&out_fds);
    if (output_ok && cb.get_current_size() > 0) {
      assert(cb.get_read_size() > 0);
      // We only care whether output is ready if the buffer is nonempty.
      FD_SET(STDOUT_FILENO, &out_fds);
    }

    int result = select(max(STDIN_FILENO, STDOUT_FILENO) + 1, 
			&in_fds, &out_fds, NULL, NULL);

    if (result < 0) {
      perror("select");

    } else if (result > 0) {

      // First, check to see if our output stream wants any data.
      if (FD_ISSET(STDOUT_FILENO, &out_fds)) {
	assert(cb.get_read_size() > 0);

	int want_chars = cb.get_read_size();
	int chars_written = 
	  write(STDOUT_FILENO, cb.read(), want_chars);
	if (chars_written < 0) {
	  if (errno != EAGAIN) {
	    perror("write");
	    exit(1);
	  }
	}

	cb.extract_characters(chars_written);

	if (_verbose) {
	  cerr << "Wrote " << chars_written << " of " << want_chars 
	       << " available; buffer " << cb.get_percent_full() 
	       << "% full\n";
	}
      }


      // Then, check to see if there's any data available on our input
      // stream.  It's possible that we've just filled up our buffer
      // from the lines above, so if that's the case, we won't bother.
      if (FD_ISSET(STDIN_FILENO, &in_fds) && 
	  cb.get_current_size() < cb.get_max_size()) {
	assert(cb.get_write_size() > 0);

	int want_chars = cb.get_write_size();
	int chars_read = 
	  read(STDIN_FILENO, cb.write(), want_chars);
	if (chars_read < 0) {
	  if (errno != EAGAIN) {
	    perror("read");
	    exit(1);
	  }
	}

	if (chars_read == 0) {
	  // End of file on input.  Drain the buffer, then exit.
	  if (_verbose) {
	    cerr << "Input fully read, draining buffer.\n";
	  }

	  // Now we want normal, blocking I/O.
	  result = fcntl(STDOUT_FILENO, F_SETFL, 0);
	  if (result < 0) {
	    perror("fcntl");
	  }

	  while (cb.get_current_size() > 0) {
	    int chars_written = 
	      write(STDOUT_FILENO, cb.read(), cb.get_read_size());
	    if (chars_written < 0) {
	      if (errno != EAGAIN) {
		perror("write");
		exit(1);
	      }
	    }
	    if (chars_written > 0) {
	      cb.extract_characters(chars_written);
	    }
	  }
	  exit(0);
	}

	cb.insert_characters(chars_read);
	if (_verbose) {
	  cerr << "Read " << chars_read << " of " << want_chars 
	       << " requested; buffer " << cb.get_percent_full() 
	       << "% full\n";
	}
      }
    }
  }
}

void BufstreamMain::
handle_opts(const string &opt, const string &arg) {
  if (opt == "b") {
    _buffer_size = atoi(arg.c_str());
    if (_buffer_size <= 0) {
      usage();
      exit(1);
    }

  } else if (opt == "f") {
    _fill_size = atoi(arg.c_str());
    if (_fill_size < 0) {
      usage();
      exit(1);
    }

  } else if (opt == "v") {
    _verbose = true;

  } else {
    MainProgram::handle_opts(opt, arg);
  }
}

void BufstreamMain::
help() {
  cerr <<
    "\n"
    "This command is a filter which buffers input from stdin and writes it to\n"
    "stdout.  It's useful for improving continuity for streaming applications\n"
    "(like soundfile players) when system I/O activity is high.\n";
  MainProgram::help();
}

void BufstreamMain::
describe_options() {
  cerr <<
    "  -b kbytes   Use a buffer of the indicated number of kilobytes.  The default\n"
    "              is " << default_buffer_kbytes << ".\n"

    "  -f kbytes   Wait for at least the indicated number of kilobytes of data to\n"
    "              be read before outputting anything at all.  The default is to\n"
    "              wait for the buffer to completely fill.\n"
    
    "  -v          Report operating status to stderr.  Primarily useful for\n"
    "              debugging.\n";

  MainProgram::describe_options();
}

void BufstreamMain::
describe_args() {
  cerr << " <infile >outfile\n\n";
}

int
main(int argc, char *argv[]) {
  BufstreamMain m;
  m.command_line(argc, argv);
  m.init();
  m.run();
  return (0);
}
